
;;; b5 expressions

;;; known expression types include:
;;; - identifiers, represented by symbols
;;; - sequences, represented by a pair whose car is 'seq and
;;;   whose cdr is a list of the contents of the sequence
;;; - explicit literals, represented by a list whose first element is
;;;   'lit and whose second element is a string
;;; - natural literals, represented by strings
;;; - an empty expression, represented by an empty string
;;; - closures, represented by a list whose elements are:
;;;   1. the symbol 'clos
;;;   2. the number of arguments the closure wants
;;;   3. a list of already given arguments, in reversed order
;;;   4. the name of the function, as a symbol
;;;   5. the body of the function
;;; - in addition, function bodies may contain formal parameters, which
;;;   are represeted by integers
;;; - results, which are represented by a list whose elements are:
;;;   1. the symbol 'res
;;;   2. the part of the expression whose value has already been
;;;      evaluated and cannot thus be modified further
;;;   3. the part of the expression that can receive arguments (a closure)
;;; - parenthesised expressions, which are represented by a pair whose
;;;   car is the symbol 'paren and whose cdr is the expression
;;; - possibly applicable closures (pacl's), which are closures followed
;;;   by something that might yet expand into a parenthesised
;;;   expression; represented by a three-element list of
;;;   1. the symbol 'pacl
;;;   2. the closure
;;;   3. the possibly-expansible expression

(define (expr-type expr)
  (cond ((pair? expr) (car expr))
	((symbol? expr) 'ident)
	((equal? expr "") 'empty)
	((string? expr) 'nlit)
	((integer? expr) 'param)
	(else (error "expr-type: unknown expression: " expr))))

(define (expr-type? expr type) (eq? type (expr-type expr)))
(define (expr-empty? expr) (expr-type? expr 'empty))

(define (expr-literal? expr) (memq (expr-type expr) '(lit nlit)))
(define (expr-result? expr) (memq (expr-type expr) '(res pres)))

(define (expr-literal-content literal)
  (if (expr-type? literal 'lit) (cadr literal) literal))

(define expr-closure-arity cadr)
(define expr-closure-arguments caddr)
(define expr-closure-name cadddr)
(define (expr-closure-body clos) (car (cddddr clos)))

(define (expr-combine e1 e2)
  (cond ((expr-empty? e1) e2)
	((expr-empty? e2) e1)
	((and (string? e1) (string? e2)) (string-append e1 e2))
	((expr-type? e2 'seq) (cons 'seq (cons e1 (cdr e2))))
	(else (list 'seq e1 e2))))

