
;;; monadic parsers
;;;
;;; a parser is(def) a function from a parsing state (incoming input +
;;; parsing position) to either
;;; - false, signifying failure to parse the requested structure
;;; - a pair of a return value and state of forthcoming parsing

;; state handlers

(define-record-type <parse-state>
  (make-state row column stream)
  state?
  (row state-row)
  (column state-column)
  (stream state-stream))

(define (state-first-elem st)
  (let ((stream (state-stream st)))
    (if (snull? stream) #f (scar stream))))

(define (state-rest-elems st)
  (let ((stream (state-stream st)))
    (if (snull? stream) stream (scdr stream))))

(define (state-advance-char st)
  (let ((rest (state-rest-elems st)))
    (case (state-first-elem st)
      ((#f) (make-state (state-row st) (state-column st) rest))
      ((#\newline) (make-state (+ (state-row st) 1) 0 rest))
      (else (make-state (state-row st) (+ (state-column st) 1) rest)))))

;; runs a parser on some input stream

(define (run-parser p stream)
  (error-init-values)
  (cond ((p (make-state 1 0 stream))
	 => (lambda (res) (car res)))
	(else (report-parse-error stream))))

;; error handling and reporting

(define error-row 1)
(define error-column 0)
(define (error-init-values)
  (set! error-row 1) (set! error-column 0))
(define (error-mark st)
  (set! error-row (max error-row (state-row st)))
  (set! error-column (max error-column (state-column st)))
  #f)

(define (report-parse-error stream)
  (let ((error-line (stream-get-line error-row stream)))
    (error (string-append
	     "Parsing error at line " (number->string error-row)
	     ", column " (number->string error-column) ": "
	     (substring error-line 0 error-column) "(*)"
	     (substring error-line error-column (string-length error-line))))))

;; basic parser combinators

; combines two parsers to an unimaginably large whole
(define (parser-bind p1 fp2)
  (lambda (st1)
    (let ((res (p1 st1)))
      (and res
	   ((fp2 (car res)) (cdr res))))))

; null-action parser for returning values
(define (parser-return val)
  (lambda (st) (cons val st)))

; Run parsers in order until one fails.
; If none fails, return a list of their results.
(define (parser-seq . ps)
  (if (null? ps) (parser-return '())
    (parser-bind
      (car ps)
      (lambda (head)
	(parser-bind
	  (apply parser-seq (cdr ps))
	  (lambda (tail)
	    (parser-return (cons head tail))))))))

; same as the former routine, except give the results as args to combine.
(define (parser-seq-combine combine . ps)
  (parser-bind
    (apply parser-seq ps)
    (lambda (result)
      (parser-return (apply combine result)))))

; run parsers in order until one succeeds.
; return the value of that parser.
(define (parser-alt . ps)
  (cond ((null? ps) (lambda (st) #f))
	((null? (cdr ps)) (car ps))
	(else
	  (lambda (st)
	    (or ((car ps) st)
		((apply parser-alt (cdr ps)) st))))))

; basic consumption parser.  Reads one element from input stream and
; succeeds if it matches the right-class? predicate.
(define (parser-item right-class?)
  (lambda (st)
    (let ((elem (state-first-elem st)))
      (if (and elem (right-class? elem))
	(cons elem (state-advance-char st))
	(error-mark st)))))

; parser that matches end-of-stream
(define (parser-end-of-stream st)
  (if (snull? (state-stream st))
    (cons #t st)
    (error-mark st)))

; read an explicitly defined value, fail otherwise.
(define (parser-item-literal elem)
  (parser-item (lambda (item) (equal? item elem))))

;; sophisticated parser combinators

; run a sequence of 0..inf instances of parser.  Combine the results
; with f and init as in fold_right (reduce).
(define (parser-chain f init parser)
  (parser-alt
    ; this could be written by parser-seq-combine if we had lazy
    ; evaluation to protect the recursive call to parser-chain
    (parser-bind
      parser
      (lambda (val1)
	(parser-bind
	  (parser-chain f init parser)
	  (lambda (val2)
	    (parser-return (f val1 val2))))))
    (parser-return init)))

; 0..inf instances of elements satisfying right-class?.
(define (parser-span right-class?)
  (parser-chain cons '() (parser-item right-class?)))

; 1..inf instances of parser.  Otherwise similar to parser-chain.
(define (parser-chain1 f init parser)
  (parser-seq-combine f parser (parser-chain f init parser)))

; 1..inf instances of elements satisfying right-class?.
(define (parser-span1 right-class?)
  (parser-chain1 cons '() (parser-item right-class?)))

; the same as parser-chain above, except that the sequence is ended by
; the success of the parser close, not by the failure of sequenced parser.
(define (parser-chain-endby f close parser)
  (parser-alt
    close
    (parser-bind
      parser
      (lambda (val1)
	(parser-bind
	  (parser-chain-endby f close parser)
	  (lambda (val2)
	    (parser-return (f val1 val2))))))))

; a sequence of literal values, given as seq.
(define (parser-seq-literal seq)
  (if (null? seq) (parser-return '())
    (parser-bind
      (parser-item-literal (car seq))
      (lambda (val1)
	(parser-bind
	  (parser-seq-literal (cdr seq))
	  (lambda (val2)
	    (parser-return (cons val1 val2))))))))

; 1..inf instances of parser, separated by sep (a parser).
(define (parser-sep-chain1 f init sep parser)
  (parser-seq-combine f
    parser
    (parser-chain
      (lambda (val1 val2) (f (cadr val1) val2))
      init
      (parser-seq sep parser))))

; 0..inf instances of parser, separated by sep (a parser).
(define (parser-sep-chain f init sep parser)
  (parser-alt
    (parser-sep-chain1 f init sep parser)
    (parser-return init)))

